/*
 * Generated by CMake build tool.
 * To modify version, open CMakeLists.txt and bump the version in project() function
 */
#pragma once

namespace dai
{
namespace build
{
    constexpr static const char* VERSION = "2.15.3";
    constexpr static const int VERSION_MAJOR = 2;
    constexpr static const int VERSION_MINOR = 15;
    constexpr static const int VERSION_PATCH = 3;

    constexpr static const char* COMMIT = "abfbd6ae78f6b9fbee34bf88ecaf4e659bd0976a";
    constexpr static const char* COMMIT_DATETIME = "2022-04-22 02:20:43 +0300";
    constexpr static const char* BUILD_DATETIME = "2022-04-22 08:35:05 +0000";
} // namespace build
} // namespace dai
